'''
Created on Oct 31, 2014

@author: ajn
'''

from __future__ import print_function
import numpy
import math
import pyparticleest.interfaces
import scipy.io
import pyparticleest.paramest.paramest as param_est
import pyparticleest.paramest.interfaces as pestint
import matplotlib.pyplot as plt
import pyparticleest.utils.kalman as kalman

class ODEPoly(pestint.ParamEst,
             pyparticleest.interfaces.FFProposeFromMeasure):
    def __init__(self, order, r, Ts, q0=100, **kwargs):
        # Define all model variables
        self.R = r * numpy.eye(1)
        #self.R = 0.01 * 2 * numpy.eye(1)
        self.params = numpy.zeros(((order + 1) ** 2 + 1,))
        self.Q = numpy.asarray(q0).reshape((1, 1))
        self.params[0] = q0
        self.Ts = Ts
        self.order = order
        super(ODEPoly, self).__init__(*kwargs)

    def propose_from_y(self, N, y, t):
        return numpy.random.normal(loc=y, scale=numpy.sqrt(self.R), size=(N, 1))

    def create_initial_estimate(self, N):
        #return numpy.random.normal(scale=2.0, size=(N, 2))
        return numpy.random.normal(loc=-0.5, scale=0.01, size=(N, 1))

    def sample_process_noise(self, particles, u, t):
        N = len(particles)
        return numpy.random.normal(scale=numpy.sqrt(self.Q), size=(N, 1))

    def measure_full(self, particles, traj, ancestors, y, t):
        return kalman.lognormpdf_scalar(y - particles.ravel(), self.R)
        #return numpy.zeros((len(particles,)))

    def sample_smooth(self, part, ptraj, anc, future_trajs, find, ut, yt, tt, cur_ind):
        return part[:, 0].reshape((-1, 1))

    def logp_xnext_full(self, part, past_trajs, pind, future_trajs, find,
                        ut, yt, tt, cur_ind):

        p = self.params[1:]

        N = len(part)
        lpx = numpy.zeros((N,))

        pcurr = part.ravel()
        pnext = future_trajs[0].pa.part[find].ravel()
        vcurr = (pnext - pcurr) / self.Ts

        if (len(past_trajs) > 0):
            anc = pind
            pold = past_trajs[-1].pa.part[anc].ravel()
            vold = (pcurr - pold) / self.Ts

            # Calculate new value of vcurr
            aold = (vcurr - vold) / self.Ts

            xvec = numpy.empty((N, (self.order + 1) ** 2))

            for i in xrange(self.order + 1):
                for j in xrange(self.order + 1):
                    xvec[:, i * (self.order + 1) + j] = (pold ** i) * (vold ** j)
#            aold_pred = (p[0] +
#             p[1] * pold + p[2] * vold +
#             p[3] * pold * pold + p[4] * pold * vold + p[5] * vold * vold +
#             p[6] * pold * pold * vold + p[7] * pold * vold * vold +
#             p[8] * pold * pold * vold * vold)
            aold_pred = xvec.dot(p).ravel()
            diffold = aold - aold_pred
            lpx += kalman.lognormpdf_scalar(diffold, self.Q)

        if (len(future_trajs) > 1):
            pnext2 = future_trajs[1].pa.part[find].ravel()
            vnext = (pnext2 - pnext) / self.Ts
            acurr = (vnext - vcurr) / self.Ts

            xvec = numpy.empty((N, (self.order + 1) ** 2))

            for i in xrange(self.order + 1):
                for j in xrange(self.order + 1):
                    xvec[:, i * (self.order + 1) + j] = (pcurr ** i) * (vcurr ** j)

#            acurr_pred = (p[0] +
#                          p[1] * pcurr + p[2] * vcurr +
#                          p[3] * pcurr * pcurr + p[4] * pcurr * vcurr + p[5] * vcurr * vcurr +
#                          p[6] * pcurr * pcurr * vcurr + p[7] * pcurr * vcurr * vcurr +
#                          p[8] * pcurr * pcurr * vcurr * vcurr)

            acurr_pred = xvec.dot(p).ravel()
            diffcurr = acurr - acurr_pred

            lpx += kalman.lognormpdf_scalar(diffcurr, self.Q)

        return lpx

    def logp_xnext_singlestep(self, part, past_trajs, pind, future_parts, find, ut, yt, tt, cur_ind):
        p = self.params[1:]

        N = len(part)
        lpx = numpy.zeros((N,))

        pcurr = part.ravel()
        pnext = future_parts[find].ravel()
        vcurr = (pnext - pcurr) / self.Ts

        if (len(past_trajs) > 0):
            anc = pind
            pold = past_trajs[-1].pa.part[anc].ravel()
            vold = (pcurr - pold) / self.Ts

            # Calculate new value of vcurr
            aold = (vcurr - vold) / self.Ts

            xvec = numpy.empty((N, (self.order + 1) ** 2))

            for i in xrange(self.order + 1):
                for j in xrange(self.order + 1):
                    xvec[:, i * (self.order + 1) + j] = (pold ** i) * (vold ** j)
#            aold_pred = (p[0] +
#             p[1] * pold + p[2] * vold +
#             p[3] * pold * pold + p[4] * pold * vold + p[5] * vold * vold +
#             p[6] * pold * pold * vold + p[7] * pold * vold * vold +
#             p[8] * pold * pold * vold * vold)
            aold_pred = xvec.dot(p).ravel()
            diffold = aold - aold_pred
            lpx += kalman.lognormpdf_scalar(diffold, self.Q)
        return lpx

    def logp_xnext_max_full(self, ptraj, uvec, yvec, tvec, cur_ind):
        return kalman.lognormpdf_scalar(numpy.zeros((1,)), self.Q)[0]

    def set_params(self, params):
        self.params = numpy.copy(params)
        self.Q = params[0:1].reshape((1, 1))

    def eval_logp_x0(self, particles, t):
        N = len(particles)
        return numpy.ones((N,))

    def maximize(self, straj):
        M = straj.traj.shape[1]

        pos = straj.traj
        vel = numpy.diff(straj.traj, axis=0) / self.Ts
        acc = numpy.diff(vel, axis=0) / self.Ts

        count = len(straj.traj) - 2
        obs = numpy.empty((count, M, (self.order + 1) ** 2))
        for i in xrange(self.order + 1):
                for j in xrange(self.order + 1):
                    obs[:, :, i * (self.order + 1) + j] = ((pos[:-2, :, 0] ** i) *
                                                            (vel[:-1, :, 0] ** j))

        res = numpy.linalg.lstsq(obs.reshape((-1, (self.order + 1) ** 2)), acc.ravel())
        err = obs.reshape((-1, (self.order + 1) ** 2)).dot(
                  res[0].reshape(((self.order + 1) ** 2, 1))).ravel() - acc.ravel()
        Q = numpy.mean(err.ravel() ** 2)
        newparams = numpy.copy(self.params)
        newparams[1:] = res[0]
        newparams[0] = Q
        return newparams

    def maximize_weighted(self, straj, alltrajs, weights):
        M = alltrajs.shape[1]

        pos = alltrajs
        vel = numpy.diff(alltrajs, axis=0) / self.Ts
        acc = numpy.diff(vel, axis=0) / self.Ts

        count = len(alltrajs) - 2
        obs = numpy.empty((count, M, (self.order + 1) ** 2))
        for i in xrange(self.order + 1):
                for j in xrange(self.order + 1):
                    obs[:, :, i * (self.order + 1) + j] = ((pos[:-2, :, 0] ** i) *
                                                            (vel[:-1, :, 0] ** j))

        for j in xrange(len(weights)):
            acc[:, j] *= math.sqrt(weights[j])
            obs[:, j] *= math.sqrt(weights[j])

        res = numpy.linalg.lstsq(obs.reshape((-1, (self.order + 1) ** 2)), acc.ravel())

        werr = numpy.empty((count, M))
        for j in xrange(len(weights)):
            err = obs[:, j].reshape((-1, (self.order + 1) ** 2)).dot(
                  res[0].reshape(((self.order + 1) ** 2, 1))).ravel() - acc[:, j].ravel()
            werr[:, j] = weights[j] * err ** 2

        Q = numpy.mean(numpy.sum(werr, axis=1))
        newparams = numpy.copy(self.params)
        newparams[1:] = res[0]
        newparams[0] = Q
        return newparams

    def copy_ind(self, particles, new_ind=None):

        if (new_ind != None):
            return numpy.copy(particles[new_ind])
        else:
            return numpy.copy(particles)



if __name__ == '__main__':
    numpy.random.seed(0)
    data = scipy.io.loadmat("vanderpol_snr43.mat")
    y = data['z']
    xtrue = data['y']

    num = 20
    max_iter = 50

    order = 2
    r = 0.01 ** 2
    #r = 0.25 ** 2
    Ts = 0.025

    theta_true = numpy.asarray((0.0, 2.0, -1.0, -2.0))

    def alpha_gen(it):
        offset = 5
        if (it <= offset):
            return 1
        else :
            return (it - offset) ** (-0.9)

    def sim(p, v, T, params):
        res = numpy.empty((T, 2))
        res[0, 0] = p
        res[0, 1] = v
        for i in xrange(1, T):
            pn = p + v * Ts
            a = 0.0
            for j in xrange(order + 1):
                for k in xrange(order + 1):
                    a += params[j * (order + 1) + k] * (p ** j) * (v ** k)
            p = pn
            v += Ts * a
            res[i, 0] = p
            res[i, 1] = v
        return res

    def callback_sim(estimator):
        plt.figure(2)
        plt.clf()

        plt.plot(estimator.straj.y[:-1], numpy.diff(estimator.straj.y) / 0.025, 'rx')

        sest = estimator.get_smoothed_estimates()

        y = sest[:-1, :, 0]
        yp = numpy.diff(sest, axis=0)[:, :, 0] / 0.025
        plt.plot(y, yp, 'g-')

        plt.plot(xtrue[:, 0], xtrue[:, 1], 'b-')

        plt.figure(5)
        plt.clf()
        plt.plot(xtrue[:, 0], xtrue[:, 1], 'b-')
        T = len(xtrue)
        p0 = xtrue[0, 0]
        v0 = xtrue[0, 1]
        pred = sim(p0, v0, T, estimator.model.params[1:])
        plt.plot(pred[:, 0], pred[:, 1], 'g--')
#        t = range(T)
#        plt.plot(t, xtrue[:, 0], 'r-')
#        plt.plot(t, pred[:, 0], 'g--')
#        plt.plot(t, xtrue[:, 1], 'b-')
#        plt.plot(t, pred[:, 1], 'c--')

        plt.draw()
        plt.show()


    model = ODEPoly(order=order, r=r, Ts=Ts, q0=100)
    estimator = param_est.ParamEstimationPSAEM(model, u=None, y=y)
    #estimator = param_est.ParamEstimation(model, u=None, y=y)

    params_it = numpy.zeros((max_iter + 1, len(model.params)))
    Q_it = numpy.zeros((max_iter + 1))
    it = 0

    def callback(params, Q, cur_iter):
        global it
        print("%04d: " % (it,), end='')
        print("Q=%6.2f " % (params[0],) , end='')
        for i in xrange(1, len(params) - 1):
            print("%6.2f " % (params[i],), end='')

        print("%6.2f" % (params[-1],))
        params_it[it, :] = params
        Q_it[it] = Q
        it = it + 1

        plt.figure(1)
        plt.clf()
        for i in xrange(1, len(theta_true)):
            plt.plot((0.0, it), (theta_true[i], theta_true[i]), 'k--')

        for i in xrange(1, len(params)):
            plt.plot(range(it), params_it[:it, i], '-')

        plt.figure(3)
        plt.clf()
        plt.plot(range(it), params_it[:it, 0], 'b-')

        plt.show()
        plt.draw()


        return (cur_iter >= max_iter)

    plt.ion()
    callback(model.params, None, 0)
    estimator.simulate(100, 1, filter='pfy', smoother='full', meas_first=True)
    callback_sim(estimator)

    ctraj = numpy.copy(estimator.straj.traj)
    callback_sim(estimator)

    (params, _) = estimator.maximize(model.params, num,
                                     filter='cpfyas',
                                     filter_options={'cond_traj': ctraj},
                                     meas_first=True,
                                     max_iter=max_iter,
                                     callback=callback,
                                     callback_sim=callback_sim,
                                     alpha_gen=alpha_gen)
    plt.ioff()
    callback_sim(estimator)

    T = len(xtrue)
    p0 = xtrue[0, 0]
    v0 = xtrue[0, 1]
    pred = sim(p0, v0, T, estimator.model.params[1:])

    scipy.io.savemat("ode_psaem_result.mat", {'params_it': params_it,
                                              'pred': pred,
                                              'xtrue': xtrue,
                                              'y': y})


